---
title: "AI for healthcare"
author: "Jean Feydy"
institute: ""
topic: ""
theme: "metropolis"
colortheme:
fonttheme:
fontsize: 11pt
linkcolor: black
urlcolor: blue
linkstyle: bold
aspectratio: 169
titlegraphic: 
logo:
date:
section-titles: false
toc: false
---


## Towards a continuous analysis of large datasets `~\cite{peyre2011numerical,heart_raw}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=35%}
`\begin{center}`{=latex}
![](images/6n-graf.pdf){height=3cm}

 \
**Simple** graph.

`\end{center}`{=latex}
::::
:::: {.column width=30%}
`\begin{center}`{=latex}
![](images/knn_graph.png){height=3cm}

 \
Manifold **hypothesis**.

`\end{center}`{=latex}
::::
:::: {.column width=35%}
`\begin{center}`{=latex}
![](images/mesh.png){height=3cm}

 \
**Physical** manifold.

`\end{center}`{=latex}
::::
:::



## Long history in physics `~\cite{solar_system,ideal_gas,nabizadeh2022covector}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=35%}
`\begin{center}`{=latex}
![](images/solar_system.png){height=3.5cm}

 \
The **Solar** system.

`\end{center}`{=latex}
::::
:::: {.column width=40%}
`\begin{center}`{=latex}
![](images/ideal_gas.png){height=3.5cm}

 \
The **ideal gas** model.

`\end{center}`{=latex}
::::
:::: {.column width=25%}
`\begin{center}`{=latex}
![](images/fluid_simulation.png){height=3.5cm}

 \
**Fluid** simulation.

`\end{center}`{=latex}
::::
:::

`\begin{center}`{=latex}

 \
Research in **physics** $~~\Longleftrightarrow~~$ **High Performance Supercomputers** \
Only available through large $\BR{institutional centers}$.

`\end{center}`{=latex}



## Recent history around video games and movies

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=33%}
`\begin{center}`{=latex}
![](images/FF.png){height=3.5cm}

 \
FFVII on the PS1 -- 1997.

`\end{center}`{=latex}
::::
:::: {.column width=33%}
`\begin{center}`{=latex}
![](images/FF_remaster.jpg){height=3.5cm}

 \
FFVII on the PS4 -- 2020.

`\end{center}`{=latex}
::::
:::: {.column width=33%}
`\begin{center}`{=latex}
![](images/jensen_huang.jpg){height=3.5cm}

 \
Jensen Huang -- 2022.

`\end{center}`{=latex}
::::
:::

`\begin{center}`{=latex}

 \
Research in **graphics** $~~\Longleftrightarrow~~$ **Graphics Processing Units** \
$\BR{Affordable}$ to any researcher: game-changer.

`\end{center}`{=latex}


## The "AI revolution" is primarily driven by hardware

Statistics and Machine Learning have been around for **decades**. \
**Breakthrough** in 2010-15: hacking $\BR{PlayStations}$ for $\BR{science}$ became **easy**.

As AI researchers, we must understand:

**1. What is a GPU?**

  - Thousands of cores, complex $\BR{memory}$ management.
  - 4 rules of GPU programming.

**2. Current trends in the semiconductor industry**

  - Just-in-time compilation, custom AI chips.
  - $\BR{Supply chain}$ issues and their impact on our careers.

## Accessible references


Coming from a **math background**:

- Chapter 2 of my PhD thesis, *Geometric data analysis, beyond convolutions*.
- Albert Chern's lecture notes at UCSD, *Introduction to computer graphics*.

Two **YouTube channels** to learn about $\BR{hardware}$:

- *Branch Education* -- to understand the circuits.
- *Asianometry* -- to get some context on the industry.

**Great** $\BR{software documentation}$ -- the source of Nvidia's monopoly in research:

- Mark Harris' posts on the Nvidia dev blog, GPU Gems textbooks.
- CUDA toolkit documentation, CUTLASS, CUB.


# What is a GPU?

## Nvidia focuses its marketing on economies of scale

`\begin{center}`{=latex}
![](images/gpu_v_cpu.jpg){height=6.5cm}

Mythbusters Demo GPU versus CPU -- 2009.

`\end{center}`{=latex}


## Nvidia focuses its marketing on economies of scale

`\begin{center}`{=latex}
![](images/gpu_v_cpu_graph.png){height=5.5cm}

Simple message: $~~~~$ 10,000 cores $~~\Longrightarrow~~$ x1,000 acceleration vs. a 10-core CPU. \
But **how did we** $\BR{fill}$ **those tubes** with the correct paintballs?

`\end{center}`{=latex}



## Scientific programs are memory bound `~\cite{traffic_jam}`{=latex}

::: columns
:::: {.column width=40%}
`\begin{center}`{=latex}
![](images/traffic_jam.jpg){height=5cm}

 \
The curse of parallelism: $\BR{traffic jams.}$

`\end{center}`{=latex}
::::
:::: {.column width=60%}
`\begin{center}`{=latex}
![](images/cable_management.jpg){height=5cm}

 \
**Structure** is required. Design **choices** \
favor $\BR{``bankable''}$ program architectures.

`\end{center}`{=latex}
::::
:::


## Let's open up a GPU

`\begin{center}`{=latex}
![](images/RTX.jpg){height=5.5cm}

`\end{center}`{=latex}



## Let's open up a GPU

`\vspace*{.25cm}`{=latex}

::: columns
:::: {.column width=50%}
`\begin{center}`{=latex}

![](images/ampere.jpg){height=5cm}

 \
**7,000 cores** on a single GPU.

`\end{center}`{=latex}
::::
:::: {.column width=50%}
`\begin{center}`{=latex}
![](images/Turing.jpg){height=5cm}

 \
The Turing **architecture**.

`\end{center}`{=latex}
::::
:::


## GPUs and large administrations follow the same plan

`\vspace*{.25cm}`{=latex}

::: columns
:::: {.column width=50%}
`\begin{center}`{=latex}

![](images/war_offices_2.png){height=5cm}

 \
GPU $~\simeq~$ **100 redundant blocks**.

`\end{center}`{=latex}
::::
:::: {.column width=50%}
`\begin{center}`{=latex}
![](images/belfast_office_2.png){height=5cm}

 \
Inside a CUDA block: **workers and buffers**.

`\end{center}`{=latex}
::::
:::



## Redundancy is key to recover high yields in spite of defects `~~\cite{dornberger1997prediction,silicon_wafer}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=50%}
`\begin{center}`{=latex}

![](images/silicon_crystal.png){height=5cm}

Silicon crystal.

`\end{center}`{=latex}
::::
:::: {.column width=50%}
`\begin{center}`{=latex}
![](images/silicon_wafer.jpg){height=5cm}

Chips are **etched onto silicon wafers**.

`\end{center}`{=latex}
::::
:::

`\begin{center}`{=latex}
GeForce RTX **3090** $~>~$ GeForce RTX **3080** $~>~$ GeForce RTX **3070** $~>~$ ...
`\end{center}`{=latex}


## GPUs are optimized to render 3D meshes in real time `~\cite{ray_tracing,Shirley2020RTW1}`{=latex}

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=50%}
`\begin{center}`{=latex}

![](images/Ray_trace_diagram.pdf){height=5cm}

Simulating light rays.

`\end{center}`{=latex}
::::
:::: {.column width=50%}
`\begin{center}`{=latex}
![](images/ray_tracing.jpg){height=5cm}

*Ray tracing in one weekend.*

`\end{center}`{=latex}
::::
:::

`\begin{center}`{=latex}
Nvidia GeForce **RTX** (**Ray Tracing** Texel eXtreme) \
$~\Longleftrightarrow~$ **Geometric** computations $+$ **textures**, on independent **patches** of the screen.
`\end{center}`{=latex}


## 5 main layers of memory storage

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=70%}

1 GPU $~\simeq~$ **100 blocks of 100 cores**.

 \
On the CPU host:

- **HDD / SSD** -- 1 TB.
- **Host RAM** -- 100 GB.

 \
On the GPU device:

- **Device RAM** -- 10 GB.
- $\BR{Shared block-wise memories}$ -- **1 Kb/core**.
- $\BR{Thread-wise registers}$ -- **1 Kb/core**.

 \
**Time**(Device RAM $\leftrightarrow$ Core) $~\simeq~$ **100 arithmetic operations.**

::::
:::: {.column width=30%}
`\begin{center}\vspace*{-.5cm}`{=latex}
![](images/memory_layers.pdf){height=7cm}

`\end{center}`{=latex}
::::
:::


## 4 rules of GPU programming

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=70%}


 \
 \
 \
**1.** Promote **block-wise** parallelism. \
\
**2.** Reduce **Host** $\leftrightarrow$ **Device** memory transfers. \
\
**3.** Reduce $\BR{Device}$  $\leftrightarrow$ $\BR{Shared/Thread}$ memory transfers. \
\
**4.** Promote **block-wise,** $\BR{contiguous}$ **memory accesses**. \

::::
:::: {.column width=30%}
`\begin{center}\vspace*{-.5cm}`{=latex}
![](images/memory_layers.pdf){height=7cm}

`\end{center}`{=latex}
::::
:::

```{=latex}
\end{frame}

\begin{frame}[fragile]{The CUDA toolkit -- a C++ dialect for GPU programming}

\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,firstnumber=8,stripnl=false,fontsize=\footnotesize]{cuda}
__global__ void 
My_CUDA_kernel(int param, float *device_data, float *device_output) {

    // We use the indices of the current thread and CUDA block
    // to assign each worker to its place in the computation plan:
    int i = blockIdx.x * blockDim.x + threadIdx.x;

    // We declare local variables as in standard C++.
    // They'll be stored in the Thread memory whenever possible:
    float some_value = 0;
    // We access the Shared memory through a raw C++ pointer:
    extern __shared__ float shared_mem[];

    // We handle transfers with a transparent interface:
    some_value    = device_data[i];  // Thread memory <- Device RAM
    shared_mem[i] = device_data[i];  // Shared memory <- Device RAM

\end{minted}

\end{frame}



\begin{frame}[fragile]{The CUDA toolkit -- a C++ dialect for GPU programming}

\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,firstnumber=8,stripnl=false,fontsize=\footnotesize]{cuda}
    

    // Computations are written in standard C++ and executed in parallel
    // by all the threads of the CUDA block:
    for(int k = 0; k < param; k++) {
        some_value = some_value + k * shared_mem[i];
        ...
    }
    
    // We may create checkpoints for all threads in a CUDA block. 
    // This may impact performances.
    __syncthreads();


    // We write results back to the Device RAM with:
    device_output[i] = some_value;  // Device RAM <- Thread memory
}
\end{minted}

\end{frame}



\begin{frame}[fragile]{The CUDA toolkit -- a C++ dialect for GPU programming}

\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,firstnumber=8,stripnl=false,fontsize=\footnotesize]{cuda}
// The main C++ program, executed by the CPU:
int main(void) {
    int N = 1024; float *host_data, *host_out, *device_data, *device_out;

    // Allocate memory on the device - the API is a bit heavy:
    cudaMalloc((void**) &device_data,  N * sizeof(float));

    // Device RAM <- Host RAM:
    cudaMemcpy(device_data, host_data, N * sizeof(float), 
               cudaMemcpyHostToDevice);

    // Set the parameters of the CUDA block: 
    int block_size = 128; int grid_size  = N / block_size;
    int shared_mem_size = 2 * block_size * sizeof(float);
    // Run the GPU kernel:
    My_CUDA_kernel<<<grid_size, block_size, shared_mem_size>>>(...);

\end{minted}

\end{frame}



\begin{frame}[fragile]{The CUDA toolkit -- a C++ dialect for GPU programming}

\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,firstnumber=8,stripnl=false,fontsize=\footnotesize]{cuda}

    // Wait for the GPU to finish its computations:
    cudaDeviceSynchronize(); 

    // Host RAM <- Device RAM:
    cudaMemcpy(host_out, device_out, N * sizeof(float),
               cudaMemcpyDeviceToHost);

    // Process and save the result "output array":
    ... 

    // Don't forget to free the allocated memory:
    cudaFree(device_data);  

    // And exit gracefully:
    return 0;
}
\end{minted}

```


## Recap on GPUs

1,000 € $~=~$ 1 GPU $~=~$ 100 $\times$ 100 cores with 5 main layers of memory:

- **Large** arrays are **slow**: $~$ Memory read/write $~\gg~$ Arithmetics.
- **Fast** buffers are **small**: $~$ 1 KB $~\simeq~$ 100 float numbers per core.

`\begin{center}`{=latex}

 \
 \
To optimize the **Shared** and **Thread** memories: $\BR{C++ or Assembly.}$ \
\
Most **scientists** rely on **pre-existing libraries** of CUDA kernels \
and $\BR{never dig deeper than the GPU Device RAM.}$
`\end{center}`{=latex}


```{=latex}
\end{frame}


\begin{frame}[fragile]{A practical example: nearest neighbor search}

\begin{columns}[T]
\begin{column}{0.7\textwidth}
\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,stripnl=false,fontsize=\footnotesize]{python}

import torch
x = torch.rand(M, D)                  # (M, D)
y = torch.rand(N, D)                  # (N, D)



diff = x.view(M,1,D) - y.view(1,N,D)  # (M, N, D)
diff2 = diff ** 2                     # (M, N, D)
sqdists = diff2.sum(dim=2)            # (M, N)
indices = sqdists.argmin(dim=1)       # (M,)

\end{minted}


\begin{center}
\textbf{Bottleneck:} \\
$(\text{M}\times\text{N}\times\text{D})$ ~~\BR{CPU} operations and memory transfers. \\
~
\end{center}
\end{column}

\begin{column}{0.30\textwidth}
\begin{center}
\includegraphics[width=\textwidth,height=7cm]{images/memory_layers.pdf}

\end{center}
\end{column}
\end{columns}
\end{frame}


\begin{frame}[fragile]{A practical example: nearest neighbor search}

\begin{columns}[T]
\begin{column}{0.7\textwidth}
\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,stripnl=false,fontsize=\footnotesize]{python}

import torch
x_ = torch.rand(M, D)                 # (M, D)
y_ = torch.rand(N, D)                 # (N, D)
x = x_.cuda()                         # (M, D)
x = y_.cuda()                         # (N, D)

diff = x.view(M,1,D) - y.view(1,N,D)  # (M, N, D)
diff2 = diff ** 2                     # (M, N, D)
sqdists = diff2.sum(dim=2)            # (M, N)
indices = sqdists.argmin(dim=1)       # (M,)

\end{minted}


\begin{center}
\textbf{Bottleneck:} \\
$(\text{M}\times\text{N}\times\text{D})$ ~~\BR{Device$\leftrightarrow$Thread} memory transfers. \\
$\vphantom{()}$
\end{center}
\end{column}

\begin{column}{0.30\textwidth}
\begin{center}
\includegraphics[width=\textwidth,height=7cm]{images/memory_layers.pdf}

\end{center}
\end{column}
\end{columns}
\end{frame}


\begin{frame}[fragile]{A practical example: nearest neighbor search}

\begin{columns}[T]
\begin{column}{0.7\textwidth}
\begin{minted}[bgcolor=blue!4,mathescape,tabsize=4,stripnl=false,fontsize=\footnotesize]{python}

import torch
x = torch.rand(M, D).cuda()            # (M, D)
y = torch.rand(N, D).cuda()            # (N, D)

# Use that |x-y|^2 = |x|^2 - 2 (x.y) + |y|^2:
dots = x @ y.T                         # (M, N)
sq_y = (y ** 2).sum(dim=1)             # (N,)

sqdists = - 2 * dots + sq_y.view(1,N)  # (M, N)
indices = sqdists.argmin(dim=1)        # (M,)

\end{minted}


\begin{center}
\textbf{Bottleneck:} \\
$(\text{M}\times\text{N}\times\text{D})$ ~~\BR{GPU} computations if $\text{D} > \text{100}$, \\
$(\text{M}\times\text{N})$ ~~\BR{Device$\leftrightarrow$Thread} memory transfers otherwise. \\
\end{center}
\end{column}

\begin{column}{0.30\textwidth}
\begin{center}
\includegraphics[width=\textwidth,height=7cm]{images/memory_layers.pdf}

\end{center}
\end{column}
\end{columns}

```

## A practical example: nearest neighbor search

`\vspace*{0cm}`{=latex}

`\begin{center}`{=latex}
![](images/genred.png){height=6cm}

On-the-fly, tiled reduction: $\BR{optimal memory management}$. \
**Bottleneck:** $(\text{M}\times\text{N}\times\text{D})~~$ GPU computations.

`\end{center}`{=latex}


## Recap on nearest neighbor search


$$
\forall \, i \in [\!\,1, \mathrm{M}\,\!]\,,~~
\text{index}[\,i\,]~\gets~ \arg \min_{j=1}^{\mathrm{N}} ~ \sum_{k=1}^{\mathrm{D}}
\big( 
    x[\,i,\,k\,] - y[\,j,\,k\,] 
    \big)^2
$$

- Each **improvement** provides a $\times\text{10}$ to $\times\text{100}~$ speed-up.

- Going even further, for **structured** data:

  - $\BR{Clusterize}$ the two point clouds.
  - **Sort** them to ensure that the clusters are **contiguous** in memory. 
  - **Skip** whole **blocks** of the tiled distance matrix.

- **Standard benchmarks** (ann-benchmarks.com) and **libraries**: FAISS...












# Compilation

## Compilation is a major bottleneck in computer science

$$
\forall \, i \in [\,1, \mathrm{M}\,]\,,~~
\text{index}[\,i\,]~\gets~ \arg \min_{j=1}^{\mathrm{N}} ~ \sum_{k=1}^{\mathrm{D}}
\big( 
    x[\,i,\,k\,] - y[\,j,\,k\,] 
    \big)^2
$$

- We have seen **4-5 different strategies**, increasingly fast but complex.
- Optimal schemes for $\mathrm{M} < \text{1,000}$ look **completely different**.

`\begin{center}`{=latex}

 \
**Naive** GPU implementations are often $\BR{x100--x1,000}$ **too slow**. \
\
Reaching optimal run times is **hard**.
`\end{center}`{=latex}


## Compilation is a deep scientific problem

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}
![](images/graph_coloring.pdf){height=6.5cm}

`\end{center}`{=latex}


## Register allocation via k-coloring of the interference graph

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}
![](images/graph_compiling.pdf){height=6.5cm}

`\end{center}`{=latex}

## LLVM: a welcome consolidation `~\cite{llvm_diagram}`{=latex}

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}
![](images/LLVM.pdf){height=6.5cm}

`\end{center}`{=latex}


## Just-in-time compilation

$\BR{Dream:}$ turn **high-level** Python code into an **optimal GPU binary**.

$\BR{Reality:}$ very hard **combinatorial** problem, $\BR{task-specific heuristics}$.

Existing libraries focus on $\BR{different targets}$:

- **Shaders** for 3D meshes.
- **Convolutions** on 2D and 3D grids -- with varying filter sizes, channels...
- Fusion of **matrix multiplications** and **non-linearities** for MLPs, Transformers.


`\begin{center}`{=latex}

 \
$\Longrightarrow~$ A **critical mass** is required to attract investments. \
 \
What about $\BR{geometric ML?}$

`\end{center}`{=latex}



## Computing libraries represent most objects as tensors



::: columns

:::: {.column width=70%}

`\vspace*{.25cm}`{=latex}
`\contexte{Context.}`{=latex}
Constrained **memory accesses** on the GPU:
`\vspace*{.3cm}`{=latex}

- **Long access times** to the registers \
  penalize the use of large **dense** arrays.

- Hard-wired **contiguous** memory accesses \
  penalize the use of **sparse** matrices.


\
`\defi{Challenge.}`{=latex}
In order to reach optimal run times:
`\vspace*{.3cm}`{=latex}

- **Restrict** ourselves to operations that are supported \
  by the constructor: convolutions, FFT, etc.

- Develop new routines from scratch in C++/CUDA \
  (FAISS, KPConv...): **several months of work**.


::::

:::: {.column width=25%}

`\begin{center}`{=latex}


```{=latex}
\begin{tikzpicture}[scale=.9]
	\draw[draw=black, fill=blue!20] (0,0) rectangle ++(3,3);
	\draw[step=.15, color=gray] (0,0) grid ++(3,3);
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=3pt, anchor = mid, fill=blue!5, rounded corners=2] at (1.5,1.5) {{\normalsize M[\,i\,,\,j\,]}};
\end{tikzpicture}
```

**Dense array**

```{=latex}
\vspace*{.5cm}

\begin{tikzpicture}[scale=.9]
	\draw[draw=black, thick] (0,0) rectangle ++(3,3);
	\node[inner sep=0pt, anchor = mid] at (1.5,1.5) {{\normalsize ($\text{i}_\text{n}$, $\text{j}_\text{n}$, $\text{M}_\text{n}$)}};
	\draw[fill=blue!20] (4*.15,1*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (15*.15,1*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (2*.15,2*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (9*.15,3*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,3*.15) rectangle ++(.15,.15);
	%
	%\draw[fill=blue!20] (3*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (6*.15,4*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,6*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (14*.15,7*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (2*.15,7*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (18*.15,8*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (1*.15,9*.15) rectangle ++(.15,.15);
	%\draw[fill=blue!20] (16*.15,10*.15) rectangle ++(.15,.15);
	%
	\draw[fill=blue!20] (6*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (16*.15,13*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (4*.15,14*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (12*.15,15*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (2*.15,16*.15) rectangle ++(.15,.15);
	\draw[fill=blue!20] (10*.15,17*.15) rectangle ++(.15,.15);
\end{tikzpicture}
```

**Sparse matrix**

`\end{center}`{=latex}

::::

:::

  
## The KeOps library: efficient support for symbolic matrices


::: columns

:::: {.column width=70%}

`\solution{Solution.}`{=latex} 
KeOps -- `\textcolor{blue}{\texttt{www.kernel-operations.io}}`{=latex}: `\vspace*{.2cm}`{=latex}

- For PyTorch, NumPy, Matlab and R, on **CPU and GPU**.
- **Automatic differentiation**.
- Just-in-time **compilation** of **optimized** C++ schemes, \
  triggered for every new **reduction**: sum, min, etc.

\
If the formula "F" is simple ($\leqslant$ 100 arithmetic operations): \
$\qquad$ "100k $\times$ 100k" computation  $\rightarrow$  10ms -- 100ms, \
$\qquad$     "1M $\times$ 1M"    computation  $\rightarrow\phantom{\text{0m}}$ 1s -- 10s.

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

Hardware ceiling of $\text{10}^{\text{12}}$ operations/s. \
**$\times \textbf{10}$ to $\times \textbf{100}$ speed-up** vs standard GPU implementations \
for a wide range of problems.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}



```{=latex}
\begin{tikzpicture}[scale=.9]
	% Left border:
	\draw[draw=black, fill=red!20] (-.45,0) rectangle ++(.3,3);
	\draw[draw=black, fill=red!80] (-.45,9*.15) rectangle ++(.3,.15);
	\draw[step=.15, color=gray] (-.45,0) grid ++(.3,3);
	\draw[draw=black] (-.45,0) rectangle ++(.3,3);
	%
	% Top border:
	\draw[draw=black, fill=green!20] (0,3.15) rectangle ++(3,.3);
	\draw[draw=black, fill=green!80] (6*.15,3.15) rectangle ++(.15,.3);
	\draw[step=.15, color=gray] (0,3.15) grid ++(3,.3);
	\draw[draw=green!50!black] (0,3.15) rectangle ++(3,.3);
	%
	\draw[draw=black,thick] (0,0) rectangle ++(3,3);
	%
	\draw[->, thick, color=red!80!black] (.5*.15, 9.5*.15) -- (5*.15, 9.5*.15);
	\draw[->, thick, color=green!50!black] (6.5*.15, 19.5*.15) -- (6.5*.15, 11*.15);
	%
	\draw[fill=blue!40] (6*.15,9*.15) rectangle ++(.15,.15);
	\node[inner sep=0pt, anchor = mid, color=blue!50!black] at (1.95,1.4) {\normalsize{F(\,\textcolor{red!80!black}{$\text{x}_\text{i}$}\,,\,\textcolor{green!50!black}{$\text{y}_\text{j}$}\,)}};
\end{tikzpicture}
```

**Symbolic matric** \
Formula + data
 

`\end{center}`{=latex}
`\vspace*{.5cm}`{=latex}


- Distances `d(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Kernel `k(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Numerous transforms.


::::

:::


```{=latex}

\end{frame}

\begin{frame}[fragile]{A first example: efficient nearest neighbor search in dimension 50}

Create large point clouds using \textbf{standard PyTorch syntax}:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
import torch
N, M, D = 10**6, 10**6, 50
x = torch.rand(N, 1, D).cuda()  # (1M,  1, 50) array
y = torch.rand(1, M, D).cuda()  # ( 1, 1M, 50) array
\end{minted}

%\pause
\vspace*{-.6cm}
Turn \textbf{dense} arrays into \textbf{symbolic} matrices:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||,highlightlines={1-2}]{python}
from pykeops.torch import LazyTensor
x_i, y_j = LazyTensor(x), LazyTensor(y)
\end{minted}

%\pause
\vspace*{-.6cm}
Create a large \textbf{symbolic matrix} of squared distances:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
D_ij = ((x_i - y_j) ** 2).sum(dim=2)  # (1M, 1M) symbolic
\end{minted}

%\pause
\vspace*{-.6cm}
Use an \texttt{.argmin()} \textbf{reduction} to perform a nearest neighbor query:
\vspace*{-.3cm}
\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4,escapeinside=||]{python}
indices_i = D_ij.argmin(dim=1)  # -> standard torch tensor
\end{minted}

\end{frame}


\begin{frame}[fragile]{The KeOps library combines performance with flexibility}

\begin{center}

```

Script of the previous slide = efficient nearest neighbor query,\
**on par** with the bruteforce CUDA scheme of the **FAISS** library...\
And can be used with $\BR{any metric}$!

```{=latex}
\end{center}

\begin{minted}[bgcolor=blue!4,fontsize=\footnotesize,mathescape,tabsize=4]{python}
D_ij = ((x_i - x_j) ** 2).sum(dim=2)     # Euclidean
M_ij =  (x_i - x_j).abs().sum(dim=2)     # Manhattan
C_ij = 1 - (x_i | x_j)                   # Cosine
H_ij = D_ij / (x_i[...,0] * x_j[...,0])  # Hyperbolic
\end{minted}

\vspace*{-.4cm}
```

KeOps supports arbitrary **formulas** and **variables** with:

- **Reductions:** sum, log-sum-exp, K-min, matrix-vector product, etc.
- **Operations:** $+$, $\times$, sqrt, exp, neural networks, etc.
- **Advanced schemes:** batch processing, block sparsity, etc.
- **Automatic differentiation:** seamless integration with PyTorch.

```{=latex}
\end{frame}

\begin{frame}{KeOps lets users work with millions of points at a time}

\begin{center}
	Benchmark of a Gaussian \textbf{convolution}\\
	between \textbf{clouds of N 3D points} on a RTX 2080 Ti GPU. 
	%\vspace*{-.3cm}

	\resizebox{.85\linewidth}{!}{\input{images/benchmarks/benchmark_gaussianconv.tex}}
\end{center}
```

## KeOps is a good fit for machine learning research

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/kmeans_2D.png){height=5cm}

`\vspace*{.25cm}`{=latex}
K-Means.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/GMM.png){height=5cm}

`\vspace*{.25cm}`{=latex}
Gaussian Mixture Model.
`\end{center}`{=latex}

::::

:::

`\begin{center}\vspace*{.3cm}`{=latex}
Use **any** kernel, metric or formula **you** like!
`\end{center}`{=latex}


## KeOps is a good fit for machine learning research

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/spectral.png){height=5cm}

`\vspace*{.25cm}`{=latex}
Spectral analysis.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/WordNet.png){height=5cm}

`\vspace*{.25cm}`{=latex}
UMAP in hyperbolic space.
`\end{center}`{=latex}

::::

:::

`\begin{center}\vspace*{.21cm}`{=latex}
Use **any** kernel, metric or formula **you** like!
`\end{center}`{=latex}


## Applications to Kriging, spline, Gaussian process, kernel regression

**A standard tool for regression** `\cite{leclercq2018bayesian}`{=latex}:

`\begin{center}`{=latex}

![](images/applications/GP_regression.png){height=3.5cm}

`\end{center}`{=latex}

Under the hood, solve a **kernel linear system**:

`\vspace*{-.5cm}`{=latex}

$$ (\lambda \,\text{Id} + K_{xx}) \,a ~=~b 
\qquad\text{i.e.}\qquad
a ~\gets~(\lambda \,\text{Id} + K_{xx})^{-1} b $$

where $~~\lambda \geqslant 0~~$ et $~~(K_{xx})_{i,j} = k(x_i,x_j)~~$ is a positive definite matrix.


## Applications to Kriging, spline, Gaussian process, kernel regression

**KeOps symbolic tensors $~~(K_{xx})_{i,j} = k(x_i,x_j)~$:**

- Can be fed to **standard solvers**: SciPy, GPyTorch, etc. \

- GPytorch on the 3DRoad dataset (N = 278k, D = 3):
  `\centerline{\defi{\textbf{7h} with 8 GPUs} $~~\rightarrow~~$ \solution{\textbf{15mn} with 1 GPU.}}`{=latex} \
   \

- Provide a **fast backend for research codes**:\
  see e.g. *Kernel methods through the roof: handling **billions of points** efficiently*, \
  by G. Meanti, L. Carratino, L. Rosasco, A. Rudi (2020).

## KeOps lets researchers focus on their models, results and theorems

Some applications to **dynamical systems** `\cite{degond2008continuum,degond2017alignment}`{=latex} \
and **statistics** `\cite{CMC}`{=latex} with A. Diez, G. Clarté et P. Degond:

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/Vicsek_3D.png){height=4cm}

`\vspace*{.25cm}`{=latex}
3D Vicsek model with orientation, \
interactive demo with 2k **flyers**.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/Vicsek_2D.png){height=4cm}

`\vspace*{.25cm}`{=latex}
2D Vicsek model on the torus, \
in real-time with 100k **swimmers**.
`\end{center}`{=latex}

::::

:::



## KeOps lets researchers focus on their models, results and theorems

$\qquad\qquad\Longrightarrow~~$ Scale up to **millions/billions** of agents with Python scripts. \

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/packing.png){height=4cm}

`\vspace*{.25cm}`{=latex}
**Packing** problem in 2D \
with 10k repulsive balls.
`\end{center}`{=latex}

::::


:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/applications/CMC_hyperbolic.png){height=4cm}

`\vspace*{.25cm}`{=latex}
Collective Monte Carlo **sampling** \
on the hyperbolic Poincaré disk.
`\end{center}`{=latex}

::::

:::



## Scaling up geometric deep learning `~\cite{gainza2020deciphering,dmasif,sverrisson2022physics}`{=latex}


```{=latex}
	\vspace*{.5cm}
	\begin{center}
		%\includegraphics[width=.75\linewidth]{images/proteins/overview.pdf}
		{\scriptsize
			\def\svgwidth{.75\linewidth}\vspace{-2mm}
			\import{images/proteins/}{overview.pdf_tex}
		}
	\end{center}
    \vspace{.5cm}
    
	\centerline{
	\begin{tabular}{ccc}
		\begin{minipage}{2.5cm}
			\includegraphics[width=\linewidth]{images/proteins/nature_cover.png}
		\end{minipage}
		&{\huge $\longrightarrow$}
		&
		\begin{minipage}{5cm}
			$\times$100\,-$\times$1,000 \textbf{faster}, \BR{lighter}\\
			\phantom{.}~~~~~and fully differentiable.
		\end{minipage}
	\end{tabular}
	}
```



## Scaling up geometric deep learning and optimal transport `~\cite{shen2021accurate}`{=latex}


`\begin{center}`{=latex}

![](images/applications/lung_registration/full.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}


## Recap on compilation

- Turning scientific code into optimal binaries is **an open problem**: \
  $\longrightarrow~~$ Massive **room for improvement** on the software side. \
  $\longrightarrow~~$ $\BR{Valuable and impactful skill.}$ \
  \
- `\solution{Symbolic}`{=latex}matrices are to **geometric** ML what \
       **sparse** matrices are to **graph** processing: \
  $\longrightarrow~~$ KeOps: **x30 speed-up** vs. PyTorch, TF et JAX. \
  $\longrightarrow~~$ Useful in a wide range of settings. \
  \
- These tools open **new paths** for geometers and statisticians: \
  $\longrightarrow~~$ GPUs are more **versatile** than you think. \
  $\longrightarrow~~$ Ongoing work to provide **fast GPU backends** to researchers, \
  $\phantom{\longrightarrow}~~~$ going beyond what Google and Facebook are ready to pay for. \


# Optimized AI cores


## Nvidia Ampere architecture in-depth `~\cite{ampere_architecture}`{=latex}

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/nvidia-a100-gpu-on-sxm4.jpg){height=6cm}

NVIDIA **A100 GPU** -- the flagship AI chip as of 2020-22.

`\end{center}`{=latex}

## Nvidia Ampere architecture in-depth `~\cite{ampere_architecture}`{=latex}

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/ga100-full-gpu-128-sms.png){height=6cm}

GA100 architecture with all 128 blocks. **A100 GPU = 108** functional blocks.

`\end{center}`{=latex}


## Nvidia Ampere architecture in-depth `~\cite{ampere_architecture}`{=latex}

`\vspace*{.25cm}`{=latex}


::: columns

:::: {.column width=60%}

`\begin{center}`{=latex}

 \
"Physical" **CUDA block** or Streaming Multiprocessor: \
\

- 192 KB of **Shared** memory. \
  \

- **4 squads** of "physical threads" or warps with: \
  \

  - 64 KB of **Thread** memory.
  - **16 int-32** cores.
  - **16 float-32** cores.
  -  **8 float-64** cores.
  -  $\BR{1 Tensor core.}$


`\end{center}`{=latex}

::::


:::: {.column width=40%}

`\vspace*{-.1cm}`{=latex}
`\begin{center}`{=latex}

![](images/ampere_block.png){height=7cm}

`\end{center}`{=latex}

::::

:::


## Integer cores: handle memory addresses -- Float-32 cores: great for 3D geometry

`\vspace*{.75cm}`{=latex}

`\begin{center}`{=latex}

![](images/numerical_int32_float32.pdf){height=6cm}

`\end{center}`{=latex}



## Float-64 cores: great for physics simulation

`\vspace*{.75cm}`{=latex}

`\begin{center}`{=latex}

![](images/numerical_float64.pdf){height=6cm}

`\end{center}`{=latex}


## Tensor cores: great for CNNs and transformers

`\vspace*{.75cm}`{=latex}

`\begin{center}`{=latex}

![](images/numerical_bfloat16.pdf){height=6cm}

`\end{center}`{=latex}






## Trading speed vs. power consumption vs. versatiliy vs. manufacturing costs

`\vspace*{.25cm}`{=latex}


::: columns

:::: {.column width=60%}

`\begin{center}`{=latex}

![](images/smartphone.png){height=4cm}

 \
"How do $\BR{Smartphone CPUs}$ Work?" \
by Branch Education.

`\end{center}`{=latex}

::::


:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/google_tpu.png){height=4cm}

 \
**Tensor Processing Units**, \
by Google.
`\end{center}`{=latex}

::::

:::





## The CPU vs. GPU uncoupling occured in the early 2000's

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/compute_scaling.pdf){height=7cm}

`\end{center}`{=latex}



## Computing power available to ML researchers

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/compute_future.pdf){height=7cm}

`\end{center}`{=latex}


# Conclusion

## A geometric tour of data science

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/geometric_tour.pdf){height=7cm}

`\end{center}`{=latex}



## What is AI research about?

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/cars.pdf){height=5cm}

**Tunnel vision** on a single angle $~\Longrightarrow~$ $\BR{high risk}$ career. \
Biggest success of the 1848 **gold rush**: Levi's **blue jeans**.

`\end{center}`{=latex}


## ML research is 100% interdisciplinary -- a mind map of my own PhD experience

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/labels.png){height=7cm}

`\end{center}`{=latex}


## Research is a deeply social and diverse activity

`\vspace*{.25cm}`{=latex}

`\begin{center}`{=latex}

![](images/people_behind_labels.png){height=7cm}

`\end{center}`{=latex}


## Some early career advice

**1.** $\BR{You}$ bring more to the table than your **potential advisor**:

- **Full-time** focus on a subject $~=~$ only during your PhD.
- Your $\BR{leverage}$: show that you are skilled and **reliable**.

 \
**2.** $\BR{Tutoring time}$ + **open** research area $~\gg~$ Prestige:

- Avoid **crowded** teams and topics.
- Outstanding environments **outside** of Paris/London/Boston/SF...
- Connect in conferences and **workshops**.


## Some early career advice

**3.** Different **countries**, different **people**, different **perspectives**. \
  $~~~~~~$ Who is the $\BR{``main character''}$ of a PhD thesis?

- I believe that it should be the **student**.
- Some people think that it is the **advisor**.

 \
**4.** $\BR{Personal chemistry}$ + **general** research area $~\gg~$ Precise topic:

- A PhD that goes according to plan is a bit disappointing anyway ;-)
- **Meet** team members (including $\BR{students}$!) before signing a long-term contract.
- **Internship** $~\simeq~$ trial period, goes both ways.



## Befriend domain experts -- Find your own balance

`\vspace*{0.5cm}`{=latex}

`\begin{center}`{=latex}

![](images/final_balance.pdf){height=6cm}

`\end{center}`{=latex}


# References